# hagelin_C.py : Hagelin de type "C" (C-35, C-36, C-38, M-209, C-446, C-48)
# 6 roues maximum, 
# le module appele par les interfaces (textes, ....)
# v4 05/12/09

import string

class Wheel:
	def __init__(self,alphabet,shift):
		self.alphabet = alphabet
		self.lg = len(alphabet)
		self.pins = []
		for i in range(self.lg):
			self.pins.append( 0 )
		self.position = 0
		self.shiftWindow = shift

	def setPinsON(self,car):
		i = string.index(self.alphabet,car)
		self.pins[i] = 1

	def setPinsOFF(self,car):
		i = string.index(self.alphabet,car)
		self.pins[i] = 0
		
	def getAPins(self,car):
		i = string.index(self.alphabet,car)
		return self.pins[i]

	def turnClockwise(self):
		self.position = (self.position + 1 ) % self.lg

	def turnCounterClockwise(self):
		self.position = (self.position + self.lg - 1 ) % self.lg 

	def getLg( self):
		return len( self.alphabet )

	def getPosition(self):
		return self.position

	def setPosition(self, newPosition ):
		self.position = newPosition

	def getExternalKey(self):
		return self.alphabet[ self.position ]

	def getInternalKey(self):
		ch = ""
		for i in range(self.lg):
			if self.pins[i] :
				ch = ch + self.alphabet[i]
			else:
				ch = ch + "_"
		return ch

	def isActive(self):
		pos_active = ( self.position + self.shiftWindow ) % self.lg
		if self.pins[ pos_active ] != 0:
			return 1
		else:
			return 0

	def zero(self):
		for i in range(self.lg):
			self.pins[i] = 0

	def info(self):
		print "Lenght       : ", self.lg
		print "Alphabet     : ", self.alphabet
		print "Shitf        : ", self.shiftWindow
		print "Pins         : ", self.pins
		print "Internal Key : ", self.getInternalKey()
		print "External key : ", self.getExternalKey()
		print "Position: ", self.position, " Actif : ", self.isActive()

class DrumBar :
	def __init__(self):
		self.nbOfWheels = 6
		self.positions = []
		for i in range(self.nbOfWheels):
			self.positions.append( 0 )
		self.leftLug  = -1
		self.rightLug = -1

	def moveLeftLug(self,position):
		if self.positions[ position ] != 0 or position == 5:  # heritage ??
			return   # ou une exception ????
		if self.rightLug != -1 and self.rightLug < position :
			return
		if self.leftLug != -1:
			self.positions[ self.leftLug ] = 0
		self.leftLug = position
		self.positions[ self.leftLug ] = 2

	def moveRightLug(self,position):
		if self.positions[ position ] != 0 or position == 0:
			return
		if self.leftLug != -1 and self.leftLug > position :
			return
		if self.rightLug != -1:
			self.positions[ self.rightLug ] = 0
		self.rightLug = position
		self.positions[ self.rightLug ] = 1

	def zeroLeftLug(self):
		if self.leftLug != -1:
			self.positions[ self.leftLug ] = 0
		self.leftLug = -1

	def zeroRightLug(self):
		if self.rightLug != -1:
			self.positions[ self.rightLug ] = 0
		self.rightLug = -1

	def zero(self):
		for i in range(self.nbOfWheels):
			self.positions[ i ] = 0
		self.leftLug  = -1
		self.rightLug = -1		

	def isActive(self,pins_actifs):
		for i in range(self.nbOfWheels):
			if self.positions[ i ] != 0 and pins_actifs[ i ]:
				return 1			
		return 0
	
	def getKey(self):
		ch = ""
		if self.leftLug != -1:	
			ch = ch + str( self.leftLug + 1 )
		else: 
			ch = ch + "0"
		ch = ch + ":"
		if self.rightLug != -1:	
			ch = ch + str( self.rightLug + 1)
		else: 
			ch = ch + "0"
		return ch

	def info(self):
		print self.positions

class Hagelin:
	def __init__(self,lettre_decalage, debug):
		self.decalage = string.index(string.ascii_uppercase,
			lettre_decalage)
		self.debug = debug
		self.wheels = []
		self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVWXYZ",15 ) )
		self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVXYZ",14  ) )
		self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTUVX",13) )
		self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRSTU",12  ) )		
		self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQRS",11) )
		self.wheels.append( Wheel("ABCDEFGHIJKLMNOPQ",10  ) )
		self.nbOfWheels = len( self.wheels )
		self.nbOfBars = 29
		self.bars = []
		for i in range( self.nbOfBars ):
			self.bars.append( DrumBar() )
		self.modeCode  = 1
		self.counter  = 0

	def setDebug(self,debug):
		self.debug = debug

	def getDebug(self):
		return self.debug

	def setLetterShift(self, lettre_decalage):
		self.decalage = string.index(string.ascii_uppercase,
			lettre_decalage)

	def getLetterShift(self):
		return string.ascii_uppercase[ self.decalage ]

	def setPinsON(self,laroue,char):
		self.wheels[laroue].setPinsON(char)

	def setPinsOFF(self,laroue,char):
		self.wheels[laroue].setPinsOFF(char)

	def getPins(self,laroue,char):
		return self.wheels[laroue].getAPins(char)

	def setALeftLug(self,aBar, position):
		self.bars[ aBar ].moveLeftLug(position)

	def setARightLug(self,aBar, position):
		self.bars[ aBar ].moveRightLug(position)

	def zeroALeftLug(self,aBar):
		self.bars[ aBar ].zeroLeftLug()

	def zeroARightLug(self,aBar):
		self.bars[ aBar ].zeroRightLug()

	def stepAWheel(self, aWheel ):
		self.wheels[ aWheel ].turnClockwise()

	def step(self):
		for i in range( self.nbOfWheels ):		
			self.wheels[i].turnClockwise()
		self.counter = self.counter + 1

	def stepBack(self):
		for i in range( self.nbOfWheels ):		
			self.wheels[i].turnCounterClockwise()
		self.counter = self.counter - 1
		if ( self.counter < 0 ):
			self.counter = 999

	def rewind(self):
		while ( self.counter != 0 ):
			self.stepBack()

	def getPosition( self ):
		return self.counter

	def getKeyOfAWheel( self, aWheel ):
		return self.wheels[ aWheel ].getExternalKey()

	def getExternalKey( self ):
		ch = ""
		for i in range( self.nbOfWheels ):
			ch = ch + self.getKeyOfAWheel( i )
		return ch

	def setExternalKey( self, key ):
		for i in range( len(key) ):
			non_trouve = 1
			lettre_cle = key[i]
			lg = self.wheels[i].getLg()
			for j in range (lg):
				if lettre_cle == self.wheels[i].getExternalKey():
					non_trouve = 0					
					break
				#self.avanceRoue(i)
				self.stepAWheel(i)
			if non_trouve: 
				print "**** erreur cle externe **** "; exit

	def getPinsOfAWheel( self, aWheel ):
		return self.wheels[ aWheel ].getInternalKey()

	def getLugsOfABar( self, aBar ):
		return self.bars[ aBar ].getKey()

	def zero(self):
		for i in range( self.nbOfWheels ):		
			self.wheels[i].zero()
		for i in range( self.nbOfBars ):
			self.bars[ i ].zero()

	def zeroPins(self):
		for i in range( self.nbOfWheels ):		
			self.wheels[i].zero()
		if self.debug:
			print "ZERO PINS"

	def cipher(self,char):
		shift = 0
		index = string.index(string.ascii_uppercase,char)
		for i in range( self.nbOfBars ):
			activesPins = []
			for j in range( self.nbOfWheels ):
				if self.wheels[ j ].isActive():
					activesPins.append( 1 )
				else:
					activesPins.append( 0 )
			if self.bars[ i ].isActive( activesPins ):
				shift = shift + 1
		j = ( 26 + self.decalage - index + shift ) % 26 
		char2 = string.ascii_uppercase[ j ]
		if ( self.debug ):
		 print "Counter: %03d Key: %6s Shift: %2d Plain: %c Crypto: %c" %  \
		  (self.getPosition(), self.getExternalKey(),shift, char, char2)
		self.step()
		return char2

	def getNbOfBars(self):
		return self.nbOfBars

	def getNbOfWheels(self):
		return self.nbOfWheels

	def toggleModeCode( self):
		if self.modeCode: self.modeCode = 0
		else: self.modeCode = 1

	def toggleDebug( self ):
		if self.debug: self.debug = 0
		else: self.debug = 1

	def getModeCode( self):
		return self.modeCode

	def info(self):
		print "Mode : ", 
		if self.modeCode: print "CODE"
		else: print "DECODE"
		for i in range( self.nbOfWheels ):
			print i+1, ":", self.wheels[i].getInternalKey()
			if ( self.debug ): self.wheels[i].info()
		print "Lugs    : ",
		for i in range( self.nbOfBars ):
			if  ( i % 6 ) == 0 : print 
			print "%2d %4s " % (i+1, self.bars[i].getKey()),
		print
		if ( self.debug ):
			for i in range( self.nbOfBars ):
				if  ( i % 3 ) == 0 : print 
				print i+1, self.bars[i].positions ,
			print		
		print "Actives Pins : ", 
		for i in range( self.nbOfWheels ):
			print self.wheels[i].isActive(),
		print
		print "Counter: ", self.getPosition(), "Key     : ", self.getExternalKey()
	

#================================================
if __name__ == "__main__":
        e = Hagelin("A", 1 )

        msg = "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        crypto = ""
        cpt = 0
        for i in range(len(msg)):
                e.step()

                car = e.cipher( msg[i] )
                crypto = crypto + car
                cpt = cpt + 1
        print crypto
